//
//  ViewController.swift
//  CalibrationTest
//
//  Created by Rocky Lin on 2/25/19.
//  Copyright © 2019 rlin. All rights reserved.
//

import UIKit
import BoltsSwift
import MetaWear
import MetaWearCpp

class ViewController: UIViewController {

    var device: MetaWear!

    var signal: OpaquePointer!

    @IBOutlet weak var connectButton: UIButton!
    @IBOutlet weak var disconnectButton: UIButton!
    @IBOutlet weak var statusLabel: UILabel!
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        if device == nil {
            self.performSegue(withIdentifier: "selectDevice", sender: self)
        }
    }

    deinit {
        device?.cancelConnection()
    }

    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if let destination = segue.destination as? DevicesTableViewController {
            destination.delegate = self
        }
    }

    @IBAction func connectButtonClicked(_ sender: Any) {
        guard device != nil else {
            self.performSegue(withIdentifier: "selectDevice", sender: self)
            return
        }
        print("Connecting")
        self.statusLabel.text = "Connecting"
        connect()
    }

    @IBAction func disconnectButtonClicked(_ sender: Any) {
        guard device != nil else {
            self.performSegue(withIdentifier: "selectDevice", sender: self)
            return
        }
        self.disconnectButton.isEnabled = false
        print("Disconnecting")
        self.statusLabel.text = "Disconnecting"
        mbl_mw_debug_disconnect(device.board)
    }

    func connect() {
        device.connectAndSetup()
            .continueOnSuccessWithTask(device.apiAccessExecutor, continuation: { disconnectTask -> Task<MblMwCalibrationState> in
                disconnectTask.continueWith(.mainThread, continuation: { t in
                    if let error = t.error {
                        print("\(error.localizedDescription)")
                        self.statusLabel.text = "Disconnected \(error.localizedDescription)"
                    } else {
                        print("Disconnected")
                        self.statusLabel.text = "Disconnected"
                    }
                    self.connectButton.isEnabled = true
                })
                DispatchQueue.main.async {
                    self.connectButton.isEnabled = false
                    self.disconnectButton.isEnabled = true
                    print("Connected")
                    self.statusLabel.text = "Connected"
                }
                self.configure()
                return self.readCalibrationState()
            }).continueOnSuccessWith(device.apiAccessExecutor, continuation: { state in
                DispatchQueue.main.async {
                    print("State: \(state.accelrometer),\(state.gyroscope),\(state.magnetometer)")
                    self.statusLabel.text = "State: \(state.accelrometer),\(state.gyroscope),\(state.magnetometer)"
                }
                if min(state.accelrometer, state.gyroscope, state.magnetometer) != SENSOR_FUSION_CALIBRATION_ACCURACY_HIGH {
                    self.calibrate()
                }
            })
    }

    func configure() {
        mbl_mw_settings_set_tx_power(device.board, 4)
        mbl_mw_settings_set_connection_parameters(device.board, 7.5, 7.5, 0, 6000)
        mbl_mw_metawearboard_set_time_for_response(device.board, 1000)
        if mbl_mw_metawearboard_lookup_module(device.board, MBL_MW_MODULE_SENSOR_FUSION) != MBL_MW_MODULE_TYPE_NA {
            mbl_mw_sensor_fusion_set_mode(device.board, MBL_MW_SENSOR_FUSION_MODE_NDOF)
            mbl_mw_sensor_fusion_set_acc_range(device.board, MBL_MW_SENSOR_FUSION_ACC_RANGE_4G)
            mbl_mw_sensor_fusion_set_gyro_range(device.board, MBL_MW_SENSOR_FUSION_GYRO_RANGE_2000DPS)
            mbl_mw_sensor_fusion_write_config(device.board)
            Thread.sleep(forTimeInterval: 1)
        }
    }

    func readCalibrationState() -> Task<MblMwCalibrationState> {
        guard let signal = mbl_mw_sensor_fusion_calibration_state_data_signal(device.board), mbl_mw_datasignal_is_readable(signal) != 0 else {
            fatalError("Invalid Signal")
        }
        let completion = TaskCompletionSource<MblMwCalibrationState>()
        mbl_mw_sensor_fusion_start(device.board)
        mbl_mw_datasignal_subscribe(signal, bridgeRetained(obj: completion)) { (object, data) in
            guard let object = object else {
                fatalError("Invalid Context")
            }
            guard let data = data else {
                fatalError("Invalid Data")
            }
            let completion = bridgeTransfer(ptr: object) as TaskCompletionSource<MblMwCalibrationState>
            let state: MblMwCalibrationState = data.pointee.valueAs()
            completion.trySet(result: state)
        }
        mbl_mw_datasignal_read(signal)
        DispatchQueue.main.async {
            print("Calibrating: mbl_mw_sensor_fusion_calibration_state_data_signal")
            self.statusLabel.text = "Calibrating: mbl_mw_sensor_fusion_calibration_state_data_signal"
        }
        return completion.task
    }

    func calibrate() {
        guard let signal = mbl_mw_sensor_fusion_calibration_state_data_signal(device.board), mbl_mw_datasignal_is_readable(signal) != 0 else {
            fatalError("Invalid Signal")
        }
        self.signal = signal
        mbl_mw_datasignal_subscribe(signal, bridge(obj: self)) { (object, data) in
            guard let object = object, let data = data else {
                return
            }
            let controller = bridge(ptr: object) as ViewController
            let state: MblMwCalibrationState = data.pointee.valueAs()
            DispatchQueue.main.async {
                print("Calibrating: \(state.accelrometer),\(state.gyroscope),\(state.magnetometer)")
                controller.statusLabel.text = "Calibrating: \(state.accelrometer),\(state.gyroscope),\(state.magnetometer)"
            }
            guard min(state.accelrometer, state.gyroscope, state.magnetometer) == SENSOR_FUSION_CALIBRATION_ACCURACY_HIGH else {
                Thread.sleep(forTimeInterval: 1)
                mbl_mw_datasignal_read(controller.signal)
                return
            }
            mbl_mw_sensor_fusion_read_calibration_data(controller.device.board, bridge(obj: controller)) { (object, board, data) in
                guard let object = object, let data = data else {
                    return
                }
                let controller = bridge(ptr: object) as ViewController
                mbl_mw_sensor_fusion_write_calibration_data(board, data)
                mbl_mw_memory_free(UnsafeMutablePointer(mutating: data))
                mbl_mw_sensor_fusion_stop(board)
                DispatchQueue.main.async {
                    print("Calibrated: mbl_mw_sensor_fusion_write_calibration_data")
                    controller.statusLabel.text = "Calibrated: mbl_mw_sensor_fusion_write_calibration_data"
                }
            }
        }
        mbl_mw_sensor_fusion_start(device.board)
        mbl_mw_datasignal_read(signal)
        print("mbl_mw_sensor_fusion_calibration_state_data_signal")
    }
}

extension ViewController: DevicesTableViewControllerDelegate {

    func didSelectDevice(_ device: MetaWear) {
        self.device = device
        self.statusLabel.text = "\(device.name) Selected"
    }
}
